#include <mlvalues.h>
#include <alloc.h>
#include <memory.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "gif.h"

/* Exception bucket for GifError */
static value giferror_exn;

/* Initialization of TkError */
value install_giferror_exn(bucket) /* ML */
  value bucket;
{
  giferror_exn = Field(bucket, 0);
  register_global_root(&giferror_exn);
  return Val_unit;
}

value loadGifFile( path )
     value path;
{
  gdImagePtr ip;
  int i;
  FILE *fp;

  value res;
  Push_roots(r,9);

  /* no error detection */
  if( (fp = fopen( String_val( path ), "r" )) == NULL ){
    raise_with_string(giferror_exn, "File not exists"); 
  }
  if( (ip = gdImageCreateFromGif( fp  )) == NULL ){
    fclose ( fp );
    raise_with_string(giferror_exn, "GIF File read error"); 
  }
  fclose ( fp );

  r[0] = alloc_string( ip->sx * ip->sy ); 
  for(i=0; i< ip-> sx; i++){
    bcopy(ip->pixels[i], String_val(r[0]) + ip->sy * i, ip->sy);
  }
  r[1] = Val_int ( ip-> sx );
  r[2] = Val_int ( ip-> sy );
  r[3] = Val_int ( ip-> colorsTotal );
  r[4] = alloc_tuple( gdMaxColors );
  for(i=0; i< gdMaxColors; i++)
    Field(r[4], i) = Val_int( ip-> red[i] );
  r[5] = alloc_tuple( gdMaxColors );
  for(i=0; i< gdMaxColors; i++)
    Field(r[5], i) = Val_int( ip-> green[i] );
  r[6] = alloc_tuple( gdMaxColors );
  for(i=0; i< gdMaxColors; i++)
    Field(r[6], i) = Val_int( ip-> blue[i] );
  r[7] = alloc_tuple( gdMaxColors );
  for(i=0; i< gdMaxColors; i++)
    Field(r[7], i) = Val_bool( ip-> open[i] );
  r[8] = Val_int( ip-> transparent );

  res = alloc_tuple(9);
  for(i=0; i<9; i++)
    Field(res, i) = r[i];

  Pop_roots ();

  /* No use of original data. Kill it. */
  gdImageDestroy( ip );

  return res;
}

